module Data.String.Ubc.Parse.Struct
( Struct(..)
, parseStruct
)
where

import Control.Arrow

import Text.Parsec

import Data.Map (Map)
import Data.String.Ubc.Parse.StructVariable (StructVariable(StructVariable))
import Data.String.Ubc.Parse.ParserState (ParserState)
import Data.String.Ubc.Parse.StructScope (StructScope(..))
import Data.String.Ubc.Parse.Scope (Scope(..))

import qualified Data.String.Ubc.Parse.Language as UbcLanguage
import qualified Data.String.Ubc.Parse.StructScope as StructScope
import qualified Data.String.Ubc.Parse.ParserState as ParserState
import qualified Data.String.Ubc.Parse.Scope as Scope

import qualified Data.Map as Map

data Struct = Struct 
        { name :: String
        , memberVariables :: Map String StructVariable
        }
        deriving (Show)

parseStruct :: Monad m => ParsecT String ParserState m Struct
parseStruct = do
        _ <- UbcLanguage.reserved "struct"

        structIdentifier <- UbcLanguage.identifier
        let structScope = StructScope
                { structName = structIdentifier 
                , variables = Map.empty
                }

        modifyState (ParserState.pushScope . ScopeStruct $ structScope)

        _ <- UbcLanguage.braces (many structMember)

        structScope <- getState >>= return . Scope.expectScopeStruct . snd . ParserState.popScope

        return $ Struct (StructScope.structName structScope) (StructScope.variables structScope)

structMember :: Monad m => ParsecT String ParserState m ()
structMember = choice [ structVariableOrFunction ]

structVariableOrFunction :: Monad m => ParsecT String ParserState m ()
structVariableOrFunction = do
        (typeName, identifier) <- try $ do
                typeName <- UbcLanguage.identifier
                objectIdentifier <- UbcLanguage.identifier
                return $ (typeName, objectIdentifier)
        choice 
                [ lookAhead (char ';') *> parseVariable typeName identifier
                ] -- TODO: Functions on structs

parseVariable variableType variableName = do
        _ <- UbcLanguage.semicolon
        -- TODO: Validate type

        (_, structScope) <- getState >>= return . (second Scope.expectScopeStruct) . ParserState.popScope
        if (== Nothing) . Map.lookup variableName . StructScope.variables $ structScope
        then do
                -- variable is not yet defined on the struct
                let structScope' = StructScope.modifyVariables (Map.insert variableName (StructVariable variableName variableType)) $ structScope
                modifyState (ParserState.pushScope $ ScopeStruct structScope')
        else do
                unexpected $ "name: \"" ++ variableName ++ "\", this name is already defined"
